from atlas_sdk.client import Client, \
  DIALOG_PARSE_TOPIC, DIALOG_ASK_TOPIC, DIALOG_SHOW_TOPIC, DIALOG_TERMINATE_TOPIC, \
  CHANNEL_ASK_TOPIC, CHANNEL_SHOW_TOPIC, INTENT_TOPIC, CHANNEL_TERMINATE_TOPIC, CHANNEL_WORK_TOPIC, \
  CHANNEL_CREATED_TOPIC, CHANNEL_DESTROYED_TOPIC
import json

class AgentClient(Client):
  """AgentClient is a simple facade between an Agent and the MQTT broker.

  It defines handler for all supported operations and exposes available publications.

  """

  def __init__(self, client_id, lang, on_ask=None, on_parse=None, on_terminate=None, on_show=None):
    """Constructs a new AgentClient.

    :param client_id: ID of the client manages by this agent
    :type client_id: str
    :param lang: Lang of the agent
    :type lang: str
    :param on_ask: Handler when a skill wants to ask something
    :type on_ask: callable
    :param on_parse: Handler when a channel wants to parse a message
    :type on_parse: callable
    :param on_terminate: Handler when a skill wants to terminate the dialog
    :type on_terminate: callable
    :param on_show: Handler when a skill wants to show something
    :type on_show: callable

    """

    super(AgentClient, self).__init__(name='agent.' + client_id) # Do not pass the client_id here!

    self._lang = lang

    self.DIALOG_PARSE_TOPIC = DIALOG_PARSE_TOPIC % client_id
    self.DIALOG_ASK_TOPIC = DIALOG_ASK_TOPIC % client_id
    self.DIALOG_SHOW_TOPIC = DIALOG_SHOW_TOPIC % client_id
    self.DIALOG_TERMINATE_TOPIC = DIALOG_TERMINATE_TOPIC % client_id
    self.CHANNEL_ASK_TOPIC = CHANNEL_ASK_TOPIC % client_id
    self.CHANNEL_SHOW_TOPIC = CHANNEL_SHOW_TOPIC % client_id
    self.CHANNEL_TERMINATE_TOPIC = CHANNEL_TERMINATE_TOPIC % client_id
    self.CHANNEL_WORK_TOPIC = CHANNEL_WORK_TOPIC % client_id
    self.CHANNEL_CREATED_TOPIC = CHANNEL_CREATED_TOPIC % client_id
    self.CHANNEL_DESTROYED_TOPIC = CHANNEL_DESTROYED_TOPIC % client_id

    self.on_ask = on_ask or self.handler_not_set
    self.on_parse = on_parse or self.handler_not_set
    self.on_terminate = on_terminate or self.handler_not_set
    self.on_show = on_show or self.handler_not_set

  def on_connect(self, client, userdata, flags, rc):
    super(AgentClient, self).on_connect(client, userdata, flags, rc)

    self.subscribe_json(self.DIALOG_TERMINATE_TOPIC, self.on_terminate)
    self.subscribe_json(self.DIALOG_ASK_TOPIC, self.on_ask)
    self.subscribe_json(self.DIALOG_SHOW_TOPIC, self.on_show)
    self.subscribe_raw(self.DIALOG_PARSE_TOPIC, self.on_parse)

    self.created()

  def intent(self, intent, data):
    """Publish a message to call the intent handler.

    :param intent: Name of the intent to call
    :type intent: str
    :param data: Data to send to the intent handler
    :type data: dict

    """

    self.publish(INTENT_TOPIC % intent, json.dumps(data))

  def created(self):
    """Inform the channel that an agent has been created for it.
    """

    self.publish(self.CHANNEL_CREATED_TOPIC, json.dumps({
      'lang': self._lang,
    }))

  def destroyed(self):
    """Inform the channel that the agent has been destroyed.
    """

    self.publish(self.CHANNEL_DESTROYED_TOPIC)

  def ask(self, payload):
    """Ask a question to the channel.

    :param payload: message
    :type payload: str

    """

    self.publish(self.CHANNEL_ASK_TOPIC, payload)

  def terminate(self):
    """Inform the channel that atlas has stopped its work.
    """

    self.publish(self.CHANNEL_TERMINATE_TOPIC)

  def work(self):
    """Inform the channel that a skill has been called so the work has started.

    It may be used to show an activity indicator.
    """

    self.publish(self.CHANNEL_WORK_TOPIC)

  def show(self, payload):
    """Show a message to the channel.

    :param payload: message
    :type payload: str

    """

    self.publish(self.CHANNEL_SHOW_TOPIC, payload)


from .agent import Agent
from .utils import find
from atlas_sdk.broker import BrokerConfig
from .discovery import Discovery, DiscoveryConfig
from .version import __version__
from .web import Server, ServerConfig
from .interpreters import Interpreter, InterpreterConfig
from .client import AtlasClient
from .executor import Executor, ExecutorConfig
from .loader import Loader, LoaderConfig
import logging, yaml, os, configparser

BROKER_CONFIG_KEY = 'broker'
INTERPRETER_CONFIG_KEY = 'interpreter'
LOADER_CONFIG_KEY = 'loader'
DISCOVERY_CONFIG_KEY = 'discovery'
EXECUTOR_CONFIG_KEY = 'executor'
SERVER_CONFIG_KEY = 'server'
LOGS_CONFIG_KEY = 'logs'

class AtlasConfig:
  """Represents the global Atlas configuration.
  """

  def __init__(self, path):
    """Constructs a new AtlasConfig for the given yml filepath.

    :param path: Path to the YAML configuration file
    :type path: str

    """

    with open(path) as f:
      data = yaml.safe_load(f)

    os.chdir(os.path.dirname(path))

    # Broker configuration

    self.broker = BrokerConfig(**data.get(BROKER_CONFIG_KEY, {}))

    # Interpreter configuration

    self.interpreter =  InterpreterConfig(**data.get(INTERPRETER_CONFIG_KEY, {}))

    # Loader configuration

    self.loader = LoaderConfig(interpreter_config=self.interpreter, **data.get(LOADER_CONFIG_KEY, {}))

    # Discovery configuration

    self.discovery = DiscoveryConfig(**data.get(DISCOVERY_CONFIG_KEY, {}))

    # Logging level

    logs = data.get(LOGS_CONFIG_KEY, {})
    log_level = logs.get('level', 'INFO')
    logging.basicConfig(level=getattr(logging, log_level))

    # Executor config

    self.executor = ExecutorConfig(broker_config=self.broker, **data.get(EXECUTOR_CONFIG_KEY, {}))

    # Server config

    self.server = ServerConfig(broker_config=self.broker, **data.get(SERVER_CONFIG_KEY, {}))

    # Disable transitions and werkzeug logging
    logging.getLogger('transitions').setLevel(logging.WARNING)

    if not self.server.debug:
      logging.getLogger('werkzeug').setLevel(logging.WARNING)

class Atlas:
  """Entry point for this assistant system.

  Atlas manages agents (representing multiple users and devices) and dialog states.

  """

  def __init__(self, config):
    """Constructs a new Atlas engine.
    
    :param config: Atlas configuration
    :type config: AtlasConfig

    """

    self._log = logging.getLogger('atlas.core')
    self._config = config
    self._agents = []
    
    self._loader = Loader(self._config.loader)
    self._executor = Executor(self._config.executor)
    self._discovery = Discovery(self._config.discovery)
    self._server = Server(self._config.server)
    self._client = AtlasClient(
      on_create=lambda d: self.create_agent(d.get('id'), d.get('uid')),
      on_destroy=self.delete_agent
    )

  def find_agent(self, id):
    """Try to find an agent in this engine.

    :param id: Id of the agent
    :type id: str
    :rtype: Agent
    """

    return find(self._agents, lambda a: a.id == id)

  def create_agent(self, id, uid):
    """Creates a new agent attached to this engine.

    :param id: Channel id for the agent
    :type id: str
    :param uid: User ID for the agent, it determines which interpreter and envs would be loaded in the Agent
    :type uid: str

    """

    if id:
      if self.find_agent(id):
        self._log.info('Reusing existing agent %s for user %s' % (id, uid))
      else:
        self._log.info('🙌 Creating agent %s for user %s' % (id, uid))
              
        agt = Agent(
          id,
          uid,
          self._loader.interpreter_for(uid),
          self._loader.env_for(uid),
          validate_intent=lambda intent_name: self._discovery.skill_env_for_intent(intent_name)
        )

        self._agents.append(agt)

        agt.start(self._config.broker)
    else:
      self._log.error('No id defined, could not create the agent')

  def delete_agent(self, id):
    """Deletes an agent from this engine.

    :param id: Id of the agent to remove
    :type id: str

    """

    agt = self.find_agent(id)

    if agt:
      self._log.info('🗑️ Deleting agent %s' % id)
      agt.cleanup()
      self._agents.remove(agt)
    else:
      self._log.info('No agent found for %s' % id)

  def cleanup(self):
    """Cleanups this engine instance.
    """

    self._log.info('Exiting atlas %s gracefuly' % __version__)

    self._discovery.cleanup()

    for agt in self._agents:
      agt.cleanup()

    self._client.stop()
    self._executor.cleanup()

  def run(self):
    """Runs this instance!
    """

    self._log.info("""

      `.-:::::::-.`      
   `-::-..`````..-::-`   
  -::.```         `.::-  
 ::- `::-`           -::    
-::   .`              ::-   atlas v%s  Copyright (C) 2018  Julien LEICHER
::.           -:`     .::   This program comes with ABSOLUTELY NO WARRANTY.
::.    ``   `-:::.    .::   This is free software, and you are welcome to redistribute it
-::   .::-`.:::::::.`.::-   under certain conditions.
 :::-:::::::::::::--:::: 
  -:::::::::::::-:.-::-  
   `-:::::::::::::::-`   
      `.-:::::::-.`      
""" % __version__)

    self._loader.load()
    self._client.start(self._config.broker)
    self._discovery.start(self._config.broker)
    self._executor.run()
    self._server.run()

    self.cleanup()


from ..utils import generate_checksum
import os, logging

class InterpreterConfig():
  """Holds information about the interpreter configuration.
  """

  def __init__(self, type='atlas.interpreter.Interpreter', **kwargs):
    """Constructs a new interpreter config.
    
    :param type: Type of the interpreter
    :type type: str
    :param path: Path where to look for training files
    :type path: str
    :param trained_path: Path where are stored trained interpreter files
    :type trained_path: str

    """

    interpreter_parts = type.split('.')
    interpreter_klass = interpreter_parts[-1:][0]
    mod = __import__('.'.join(interpreter_parts[:-1]), fromlist=[interpreter_klass])

    self.interpreter_class = getattr(mod, interpreter_klass)
    self.kwargs = kwargs

  def construct(self):
    """Constructs a new interpreter related to this configuration instance.
    """

    return self.interpreter_class(**self.kwargs)

class Interpreter():
  """Interpreters convert natural language sentences to parsed structures used by atlas.

  There multiple open-source NLU libraries out there so it should be fairly easy to write a tiny abstraction
  for atlas.

  Interpreters should expose metadata since this is used by the Agent (ie. the dialog engine).

  """

  def __init__(self, name):
    """Constructs a new interpreter.

    :param name: Name of the interpreter
    :type name: str

    """

    self._log = logging.getLogger('atlas.interpreter.%s' % name)
    self.name = name

  def checksum_match(self, data, checksum_file_path):
    """Checks if the checksum is the same between raw data and a checksum file.

    :param data: Raw data for which we want to compute the checksum
    :type data: str
    :param checksum_file_path: Path of the checksum file
    :type checksum_file_path: str
    :rtype: tuple

    """

    data_checksum = generate_checksum(data)

    self._log.debug('Checksum of raw data is %s' % data_checksum)

    try:
      with open(checksum_file_path) as f:
        file_checksum = f.read()

      self._log.debug('File checksum is %s' % file_checksum)

      return (data_checksum == file_checksum, data_checksum)
    except FileNotFoundError:
      self._log.debug('Checksum file not found at %s' % checksum_file_path)
      return (False, data_checksum)

  def training(self):
    """Returns annotated training data used to compute accuracy. This is needed
    only if you plan to use atlas-check with this interpreter.

    You must returns an array in the following form:

    [{
      'text': 'will it rain today',
      'intent': 'weatherForecast',
    }]

    :rtype: list

    """
    
    return []

  def lang(self):
    """Returns the interpreter language.

    :rtype: str

    """

    return 'en' # Default to en

  def metadata(self):
    """Gets interpreter metadata as a dict such as
    {
        'intent_name': ['slot_name', 'another_slot']
    }

    :rtype: dict

    """

    return {}

  def fit(self, training_file_path, trained_directory_path):
    """Train this interpreter.

    :param training_file_path: Path to the training file
    :type training_file_path: str
    :param trained_directory_path: Path to the trained folder where stuff should be put
    :type trained_directory_path: str
    
    """

    pass

  def parse(self, msg):
    """Parses an incoming message and converts it to a list of dict that represents intents and slots.

    If multiple values are found for the same slot, the interpreter should returns an array of values
    instead of a single one.

    Why a list of dict? Because you can return multiple intents for the same sentence if this is what
    the user asked for.

    :param msg: Message to parse
    :type msg: str

    :rtype: list

    """

    return []

  def parse_entity(self, msg, intent, slot):
    """Parses a message as an entity. This is useful when the skill is asking for mandatory parameters.

    The interpreter should translate the given value into the appropriate entity type.

    :param msg: Raw message to parse
    :type msg: str
    :param intent: Intent requested
    :type intent: str
    :param slot: Slot requested
    :type slot: str

    :rtype: str

    """

    return msg



from atlas_sdk.broker import BrokerConfig
import os, logging, glob, subprocess

class ExecutorConfig:
  """Configuration for an Executor.
  """

  def __init__(self, path, broker_config):
    """Constructs a new ExecutorConfig.
    
    :param path: Path of the skills directory
    :type path: str
    :param broker_config: Broker configuration used to start the skills on the specific message broker
    :type broker_config: BrokerConfig

    """

    self.path = os.path.abspath(path)
    self.broker = broker_config

class Executor:
  """Executor is a handy utility used to launch skill placed in a specific directory.

  It looks for directories with an "atlas" file containing the command to run and executes it with the configurated
  broker configuration.

  """

  def __init__(self, config):
    """Constructs a new executor for the given skills folder.
    
    :param config: Executor configuration
    :type config: ExecutorConfig

    """

    self._log = logging.getLogger('atlas.executor')
    self._processes = []
    self._config = config

  def run(self):
    """Run discover skills and run them.
    """

    self._log.info('Running executor in: %s' % self._config.path)

    for p in glob.glob(self._config.path + '/**/atlas'):

      with open(p) as f:
        # Constructs the command line to run
        cmd = "%s -H %s -p %d" % (f.read(), self._config.broker.host, self._config.broker.port)

        if self._config.broker.is_secured():
          cmd += " -u %s:%s" % (self._config.broker.username, self._config.broker.password)

        process = subprocess.Popen(cmd, cwd=os.path.dirname(p), shell=True)
        self._processes.append(process)
        self._log.info('🚀 Started %s' % cmd)

  def cleanup(self):
    """Stops spawned processes.
    """

    self._log.info('Stopping processes')

    for process in self._processes:
      process.terminate()
      self._log.info('Stopped %s' % process.args)



from .interpreters import InterpreterConfig
import logging, os, configparser, sys

DEFAULT_SECTION = 'DEFAULT'
DEFAULT_KEY = 'default'

class LoaderConfig():
  """Loads interpreters and env configurations from the file system.
  """

  def __init__(self, interpreter_config, training_path, trained_path, env_path):
    """Constructs a new LoaderConfig object.

    :param interpreter_config: Configuration of the interpreter
    :type interpreter_config: InterpreterConfig
    :param training_path: Where are stored the training samples
    :type training_path: str
    :param trained_path: Where are stored the trained data (checksum + backend specific files to speed up the loading time)
    :type trained_path: str
    :param env_path: Where are stored the configuration parameters
    :type env_path: str

    """

    self.interpreter_config = interpreter_config
    self.trained_path = os.path.abspath(trained_path)
    self.training_path = os.path.abspath(training_path)
    self.env_path = os.path.abspath(env_path)

class Loader():
  """Loads interpreters and env configurations from the file system.
  """

  def __init__(self, config):
    """Instantiates a new Loader.

    :param config: Loader configuration
    :type config: LoaderConfig

    """

    self._log = logging.getLogger('atlas.loader')
    self._config = config
    self._envs = {}
    self._interpreters = {}

  def load(self):
    """Starts the loader!
    """

    self._load_envs()
    self._load_interpreters()

  def env_for(self, uid):
    """Retrieve environment variables for the given user id.

    :param uid: User id
    :type uid: str
    :rtype: dict

    """

    self._log.debug('Retrieving environment for %s' % uid)

    return self._envs.get(uid, self._envs.get(DEFAULT_KEY))

  def interpreter_for(self, uid):
    """Retrieve interpreter for the given user id.

    :param uid: User id
    :type uid: str
    :rtype: Interpreter

    """

    self._log.debug('Retrieving interpreter for %s' % uid)

    return self._interpreters.get(uid, self._interpreters.get(DEFAULT_KEY))

  def _load_envs(self):
    """Load user environments variables. Filenames should match the user ID.    
    """

    self._log.info('📝 Loading user envs from %s' % self._config.env_path)

    for env_file_path in os.listdir(self._config.env_path):
      with open(os.path.join(self._config.env_path, env_file_path)) as f:
        config_string = ('[%s]\n' % DEFAULT_SECTION) + f.read() # Add a default section since the file is a plain empty one
        config = configparser.ConfigParser()
        config.optionxform = str
        config.read_string(config_string)

        uid, _ = os.path.splitext(env_file_path)

        self._envs[uid] = dict(config[DEFAULT_SECTION])

    if DEFAULT_KEY not in self._envs:
      self._log.critical('No %s env file found, exiting' % DEFAULT_KEY)
      sys.exit(1)

    self._log.debug('Loaded %d env(s) files' % len(self._envs))

  def _load_interpreters(self,):
    """Load user interpreters. Filenames should match the user ID.
    """

    self._log.info('📝 Loading user interpreters from training files %s' % self._config.training_path)

    for training_file_path in os.listdir(self._config.training_path):
      uid, _ = os.path.splitext(training_file_path)

      interpreter = self._config.interpreter_config.construct()
      interpreter.fit(os.path.join(self._config.training_path, training_file_path), self._config.trained_path)

      self._interpreters[uid] = interpreter

    if DEFAULT_KEY not in self._interpreters:
      self._log.critical('No %s interpreter file found, exiting' % DEFAULT_KEY)
      sys.exit(1)

    self._log.debug('Loaded %d interpreter(s)' % len(self._interpreters))


from flask import Flask, request, render_template, send_file
from flask_restful import Api, Resource
from flask_socketio import SocketIO, Namespace, emit
import logging, subprocess, os
from ..version import __version__
from atlas_sdk import BrokerConfig, ChannelClient

app = Flask('atlas.web', static_folder='./public', template_folder='./public')
app.config['TEMPLATES_AUTO_RELOAD'] = True

api = Api(app)
socketio = SocketIO(app)

@app.route('/')
def index():
  # TODO language from user settings

  return render_template('index.html', version=__version__)

@app.route('/sw.js')
def serve_service_worker():
  return send_file('public/sw.js')

class ServerConfig:
  """Holds settings related to the web server.
  """

  def __init__(self, broker_config, url=None, host='localhost', port=5000, debug=True):
    """Constructs a new ServerConfig.

    :param broker_config: Broker config used by the web server for websocket channels
    :type broker_config: BrokerConfig
    :param host: Host to bind to
    :type host: str
    :param port: Port to bind to
    :type port: int
    :param debug: Determines if we should run the web & webpack watch command
    :type debug: bool

    """

    self.url = url or 'http://%s:%d' % (host, port)
    self.host = host
    self.port = port
    self.debug = debug
    self.broker = broker_config

class HelloWorld(Resource):
  def get(self):
    return {'hello': 'world'}

class Server:
  """Web server for the atlas interface.
  """

  def __init__(self, config):
    """Constructs a new Server with the given config.

    :param config: Configuration of the web server
    :type config: ServerConfig

    """

    self._log = logging.getLogger('atlas.server')
    self._channels = {}
    self._config = config

    api.add_resource(HelloWorld, '/hello')

    socketio.on_event('connect', self.on_connect)
    socketio.on_event('disconnect', self.on_disconnect)
    socketio.on_event('parse', self.on_parse)

  def on_connect(self):
    """Called when a connection has been made by a websocket.
    """

    client_id = request.sid

    channel = ChannelClient(client_id, 1337, # TODO Replace with a true user id
      on_ask=lambda d, _: socketio.emit('ask', d, room=client_id),
      on_show=lambda d, _: socketio.emit('show', d, room=client_id),
      on_terminate=lambda: socketio.emit('terminate', room=client_id),
      on_work=lambda: socketio.emit('work', room=client_id),
      on_created=lambda d, _: socketio.emit('created', d, room=client_id),
      on_destroyed=lambda: socketio.emit('destroyed', room=client_id),
    )
    
    self._channels[client_id] = channel

    channel.start(self._config.broker)

  def on_disconnect(self):
    """Called when a websocket has disconnected.
    """

    self._channels[request.sid].stop()

    del self._channels[request.sid]

  def on_parse(self, data):
    """Called when a websocket client wants to parse data.

    :param data: Data received
    :type data: str

    """

    self._channels[request.sid].parse(data)

  def run(self):
    """Starts the web server.
    """
        
    # When in a debug mode, start webpack in watch mode automatically
    # This help makes the development easier
    if self._config.debug:
      p = subprocess.Popen('npm run start', cwd=os.path.dirname(__file__), shell=True)
      self._log.info('Started webpack')

    self._log.info('🌐 Starting web server on %s:%s, public url is %s' % (self._config.host, self._config.port, self._config.url))
    # app.run(self._config.host, self._config.port)#, debug=self._config.debug)

    socketio.run(app, self._config.host, self._config.port)

    # And terminate the process once done
    if self._config.debug:
      p.terminate()
      self._log.info('Stopped webpack')


import uuid, hashlib

def generate_checksum(data):
  """Generates a checksum from a raw string.

  :param data: Data for which you want an hex string
  :type data: str
  :rtype: str

  """

  return hashlib.sha256(data.encode('utf-8')).hexdigest()

def resolve_parametric_intent_name(name, slots):
  """Format the intent name with slots dict to handle parametric intent name.

  With this feature, you can create generic intent and output a single intent name
  based on slots values.

  :param name: Name of the intent, may contain placeholders such as {slotName}
  :type name: str
  :param slots: Dictionary of slots values
  :type slots: dict

  """

  # TODO what to do if a slot value is an array?

  try:
    return name.format(**slots)
  except KeyError:
    return name

def generate_hash():
  """Generates a random hash.

  :rtype: str

  """

  return uuid.uuid4().hex

def find(iterable, predicate):
  """Find an element in a collection.

  :param iterable: Collection to iter in
  :type iterable: iter
  :param predicate: Predicate to use
  :type predicate: lambda

  """
  
  eles = list(filter(predicate, iterable))[:1]

  if eles:
    return eles[0]
  else:
    return None


from .loader import Loader
from .interpreters import Interpreter
from .atlas import AtlasConfig
import unittest, logging

class CheckerTest(unittest.TestCase):
  def __init__(self, interpreter):
    """Constructs a new test case for one interpreter.

    :param interpreter: Interpreter to test
    :type interpreter: Interpreter

    """

    super(CheckerTest, self).__init__()
    
    self._interpreter = interpreter

  def runTest(self):
    for data in self._interpreter.training():
      text = data['text']
      
      with self.subTest(text=text):
        expectedIntent = data['intent']
        r = self._interpreter.parse(text)

        self.assertEqual(1, len(r))
        self.assertEqual(expectedIntent, r[0]['intent'])

        # TODO may be add slot check support

class Checker:
  """Tiny helper to check interpreter accuracy. This makes it easy to know
  which intent may not be correctly inferred and so which one needs more training
  data.
  """

  def __init__(self, atlas_config):
    """Constructs a new checker instance.

    :param atlas_config: Atlas configuration
    :type atlas_config: AtlasConfig

    """

    self._log = logging.getLogger('atlas.checker')
    self._loader = Loader(atlas_config.loader)

  def run(self):
    """Runs the checker.
    """

    self._loader.load()

    suite = unittest.TestSuite()

    for uid, interp in self._loader._interpreters.items():
      suite.addTest(CheckerTest(interp))
      self._log.info('Added interpreter %s tests for uid %s' % (interp.name, uid))

    unittest.TextTestRunner().run(suite)


from . import Interpreter
from snips_nlu import load_resources, SnipsNLUEngine, __version__
from snips_nlu.builtin_entities import BuiltinEntityParser, is_builtin_entity
import io, json, os

def get_entity_value(entity, default_value=None):
  """Retrieve an entity value.
  
  :param entity: Dictionary wich contains results
  :type entity: dict
  :param default_value: Default value if not found
  :type default_value: any

  """

  return entity.get('value', entity.get('from', default_value))

class SnipsInterpreter(Interpreter):

  def __init__(self):
    super(SnipsInterpreter, self).__init__('snips')

    self._meta = None
    self._lang = None
    self._engine = None
    self._entity_parser = None

    self._log.info('Using snips-nlu version %s' % __version__)

  def metadata(self):
    return self._meta

  def lang(self):
    return self._lang

  def fit(self, training_file_path, trained_directory_path):
    filename, _ = os.path.splitext(os.path.basename(training_file_path))

    # TODO check what should be in the base Interpreter class

    trained_path = os.path.join(trained_directory_path, '%s.trained.json' % filename)
    checksum_path = os.path.join(trained_directory_path, '%s.checksum' % filename)

    with open(training_file_path) as f:
      training_str = f.read()
      training_data = json.loads(training_str)
      self._lang = training_data['language']
      self._log.info('Loading resources for language %s' % self._lang)
      load_resources(self._lang)

    same, computed_checksum = self.checksum_match(training_str, checksum_path)

    # Checksums match, load the engine from trained file
    if same and os.path.isfile(trained_path):
      self._log.info('Checksum matched, loading trained engine')
      with open(trained_path) as f:
        self._engine = SnipsNLUEngine.from_dict(json.load(f))  
    else:
      self._log.info('Checksum has changed, retraining the engine')
      self._engine = SnipsNLUEngine()
      self._engine.fit(training_data)

      with open(trained_path, mode='w') as f:
        json.dump(self._engine.to_dict(), f)

      with open(checksum_path, mode='w') as f:
        f.write(computed_checksum)

    self._entity_parser = BuiltinEntityParser(self._lang)
    self._meta = { k: list(v.keys()) for k, v in self._engine._dataset_metadata['slot_name_mappings'].items() }

  def parse_entity(self, msg, intent, slot):
    entity_label = self._engine._dataset_metadata['slot_name_mappings'].get(intent, {}).get(slot)

    # TODO try to find a way to retrieve multiple slot values, that's a hard one
    # May be we can try matching on _dataset_metadata['entities']

    if is_builtin_entity(entity_label):
      parsed = self._entity_parser.parse(msg)

      if parsed:
        return get_entity_value(parsed[0]['entity'], msg)

    return msg

  def parse(self, msg):

    # TODO manage multiple intents in the same sentence

    parsed = self._engine.parse(msg)

    if parsed['intent'] == None:
      return []

    slots = {}

    # Constructs a slot dictionary with slot value as a list if multiples matched
    for slot in parsed['slots']:
      name = slot['slotName']
      value = get_entity_value(slot['value'])

      if name in slots:
        if type(slots[name]) is not list:
          slots[name] = [slots[name]]

        slots[name].append(value)
      else:
        slots[name] = value

    return [{
      'text': msg,
      'intent': parsed['intent']['intentName'],
      'slots': slots,
    }]


from atlas_sdk.client import Client, DISCOVERY_PING_TOPIC, DISCOVERY_PONG_TOPIC
from ..version import __version__
import json
from datetime import datetime

class DiscoveryClient(Client):

  def __init__(self, on_discovery=None):
    """Constructs a new DiscoveryClient.

    :param on_discovery: Handler when a discovery response has been retrieved
    :type on_discovery: callable

    """

    super(DiscoveryClient, self).__init__(name='discovery')

    self.on_discovery = on_discovery or self.handler_not_set
    self._started_at = datetime.utcnow()

  def on_connect(self, client, userdata, flags, rc):
    super(DiscoveryClient, self).on_connect(client, userdata, flags, rc)

    self.subscribe_json(DISCOVERY_PONG_TOPIC, self.on_discovery)

  def ping(self):
    """Launch a ping request for the discovery service.
    """

    self.publish(DISCOVERY_PING_TOPIC, json.dumps({
      'version': __version__,
      'started_at': self._started_at.isoformat(), # TODO format it like snips-nlu
    }))


from .atlas import Atlas, AtlasConfig
from .version import __version__
from .checker import Checker
from .web import Server
from .utils import generate_hash
from atlas_sdk import ChannelClient
import os, sys, argparse
from cmd import Cmd

# Monkey patch with eventlet
try:
  import eventlet # pylint: disable=E0401
  eventlet.monkey_patch()
except:
  pass

class Prompt(Cmd):
  def __init__(self, broker_config):
    super(Prompt, self).__init__()

    self.prompt = '> '
    self.uid = 'client'
    self.client_id = generate_hash()
    self.client = ChannelClient(self.client_id, self.uid, 
      on_ask=self.show_message, 
      on_show=self.show_message, 
      on_terminate=self.has_terminated,
      on_work=self.on_work
    )
    self.client.start(broker_config)

  def show_message(self, data, raw):
    print (data.get('text'))

  def on_work(self):
    print('-- work has started')
  
  def has_terminated(self):
    print ('-- intent has terminated')

  def default(self, arg):
    self.client.parse(arg)

  def do_exit(self, args):
    self.client.stop()

    raise SystemExit

def get_config():
  parser = argparse.ArgumentParser(description='Atlas CLI %s - An open-source assistant built for people' % __version__)

  parser.add_argument('-c', '--config', help='Path to the configuration yaml file')

  args = parser.parse_args(sys.argv[1:])

  config_path = os.path.abspath(args.config or 'atlas.yml')

  return AtlasConfig(config_path)

def client():
  prompt = Prompt(get_config().broker)
  prompt.cmdloop('Welcome!')

def web():
  serv = Server(get_config().server)
  
  try:
    serv.run()
  except Exception as e:
    print (e)

def check():
  check = Checker(get_config())
  check.run()

def main():
  atlas = Atlas(get_config())

  try:
    atlas.run()
  except Exception as e:
    print (e)
    atlas.cleanup()

if __name__ == '__main__':
  main()


from . import Interpreter

class DummyInterpreter(Interpreter):
  """This is a tiny dummy interpreter used for testing.
  """

  def __init__(self):
    super(DummyInterpreter, self).__init__('dummy')
    
  def metadata(self):
    return {
        'weatherForecast': ['city', 'date'],
        'reminder': ['date', 'subject', 'message']
    }

  def training(self):
    return [{
      'text': 'will it rain today',
      'intent': 'weatherForecast'
    }, {
      'text': "what's the weather like today",
      'intent': 'weatherForecast'
    }]

  def parse(self, msg):
    return [{
      'text': msg,
      'intent': 'weatherForecast',
      'slots': {
        'date': '2018-05-20',
      }
    }]


import logging
from .version import __version__
from .client import AgentClient
from .utils import generate_hash, resolve_parametric_intent_name
from atlas_sdk import BrokerConfig
from atlas_sdk.request import SID_KEY, UID_KEY, ENV_KEY, VERSION_KEY, LANG_KEY, CID_KEY
from .interpreters import Interpreter
from transitions import Machine, EventData, MachineError
from transitions.extensions.states import add_state_features, Timeout

STATE_ASLEEP = 'asleep'
INTENT_CANCEL = 'cancel' # TODO handle the cancel event, see comments in the Agent.parse method
INTENT_NOTFOUND = 'notfound' # TODO handle the case when no intent has been found in the user request
PREFIX_ASK = 'ask__'

def to_ask_state(slot):
  """Converts to an ask state.

  :param slot: Slot to convert
  :type slot: str

  """

  return PREFIX_ASK + slot

@add_state_features(Timeout)
class AgentMachine(Machine):
  """Custom transitions Machine to add state features.
  """
  
  pass

class Agent:
  """Agents are the core of the dialog engine of atlas.

  They maintain the state of a conversation and act as a relay between skills and channels. They are
  tied to a particular user or device and open a MQTT client themselve.

  """

  def __init__(self, id, uid, interpreter, env, validate_intent=None, ask_timeout=30):
    """Creates a new agent.

    :param id: Channel id
    :type id: str
    :param uid: User ID
    :type uid: str
    :param interpreter: Interpreter to be used
    :type interpreter: Interpreter
    :param env: Configuration parameters for the user
    :type env: dict
    :param validate_intent: Handler to validate the user intent by checking if it's available, 
                            it will take the intent name and should returns a dict containing valid env 
                            variables if the skill exists or None if no skill could be found matching
                            that intent
    :type validate_intent: callable
    :param ask_timeout: Timeout when entering an ask state
    :type ask_timeout: int

    """

    self._log = logging.getLogger('atlas.agent.%s' % id)
    self._intent_queue = []

    self.validate_intent = validate_intent
    self.interpreter = interpreter
    self.env = env
    self.id = id
    self.uid = uid

    # Configure the client facade

    self._client = AgentClient(id, self.interpreter.lang(),
      on_parse=self.parse,
      on_ask=self.ask,
      on_terminate=self._terminate_from_skill,
      on_show=self.show
    )

    self.reset()

    # Constructs every possible transitions from interpreter metadata

    metadata = self.interpreter.metadata()

    ask_states = list(set([to_ask_state(slot) for meta in metadata.values() for slot in meta]))
    states = [STATE_ASLEEP] + list(metadata.keys()) + [{ 
      'name': o, 
      'timeout': ask_timeout, 
      'on_timeout': self._on_timeout 
    } for o in ask_states]

    self._machine = AgentMachine(self, 
      states=states, 
      initial=STATE_ASLEEP, 
      send_event=True, 
      before_state_change=lambda e: self._log.info('⚡ %s: %s -> %s' % (e.event.name, e.transition.source, e.transition.dest) )
    )

    self._log.info('Created with states %s' % list(self._machine.states.keys()))

    self._machine.add_transition(STATE_ASLEEP, '*', STATE_ASLEEP, after=self.reset)

    ask_transitions_source = { k: [] for k in ask_states }

    for intent, slots in metadata.items():
      converted_slots = [to_ask_state(s) for s in slots]
      self._machine.add_transition(intent, [STATE_ASLEEP] + converted_slots, intent, after=self._call_intent)

      for slot in converted_slots:
        ask_transitions_source[slot].append(intent)

    for k, v in ask_transitions_source.items():
      self._machine.add_transition(k, v, k, after=self._on_asked)

  def reset(self, event=None):
    """Resets the current agent states.

    :param event: Optional event parameters
    :type event: EventData

    """

    self._cur_asked_param = None
    self._cur_intent = None
    self._cur_conversation_id = None
    self._cur_slots = {}

    self._client.terminate()

    self._process_next_intent()

  def go(self, trigger_name, **kwargs):
    """Safely call a trigger and catch errors

    :param trigger_name: Name of the trigger
    :type trigger_name: str

    """

    try:
      self.trigger(trigger_name, **kwargs) # pylint: disable=E1101
    except MachineError as err:
      self._log.error('Could not trigger "%s": %s' % (trigger_name, err))

  def _call_intent(self, event):
    """Call the intent with current slot values.

    :param event: Machine event
    :type event: EventData

    """

    # Here we need to keep track of the original intent name otherwise transitions
    # will be broke
    self._cur_intent = event.transition.dest

    # But we use the resolved name for the intent validation to check if a skill can
    # take care of it
    resolved_intent_name = resolve_parametric_intent_name(self._cur_intent, self._cur_slots)
    valid_keys = self.env.keys()

    # Try to validate if the intent could be reached

    if self.validate_intent:
      valid_keys = self.validate_intent(resolved_intent_name)

      if valid_keys == None:
        self._log.warn('Intent %s could not be reached, skipping now' % resolved_intent_name)
        return self.go(STATE_ASLEEP)

    self._cur_conversation_id = generate_hash()
    
    self._client.work()

    # Constructs the message payload
    
    data = {
      CID_KEY: self._cur_conversation_id,
      SID_KEY: self.id,
      UID_KEY: self.uid,
      LANG_KEY: self.interpreter.lang(),
      VERSION_KEY: __version__,
      ENV_KEY: { k: self.env[k] for k in valid_keys },
    }
    
    data.update(self._cur_slots)

    self._log.debug('Calling resolved intent "%s (%s)" with params %s' % (resolved_intent_name, self._cur_intent, data))

    self._client.intent(resolved_intent_name, data)

  def _on_asked(self, event):
    """Entered in ask state, save current asked param.

    :param event: Machine event
    :type event: EventData

    """

    self._cur_asked_param = event.transition.dest[len(PREFIX_ASK):]

    payload = event.kwargs.get('payload')

    self._log.debug('Asking request with payload %s' % payload)

    self._client.ask(payload)

  def _on_timeout(self, event):
    """Called when a state timeout has been reached.

    :param event: Machine event
    :type event: EventData

    """

    self.go(STATE_ASLEEP)

  def parse(self, msg):
    """Parse a raw message.

    :param msg: Message to parse
    :type msg: str

    """

    self._log.debug('Parsing "%s"' % msg)

    # TODO if intent is "cancel", go to the cancel state immediately. This will generates a new
    # conversation id so old request will be dismissed if they tried to do something

    # Start by checking if we are in a ask* state
    if self.state.startswith(PREFIX_ASK) and self._cur_asked_param: # pylint: disable=E1101
      self._cur_slots[self._cur_asked_param] = self.interpreter.parse_entity(msg, self._cur_intent, self._cur_asked_param)

      self.go(self._cur_intent)
    else:
      data = self.interpreter.parse(msg)

      # TODO if no intent was found, let it know with the INTENT_NOTFOUND

      self._intent_queue.extend(data)

      if self.state == STATE_ASLEEP: # pylint: disable=E1101
        self._process_next_intent()

      # TODO another intent is running, ask for confirmation? maybe?

  def _process_next_intent(self):
    """Process the intent queue if any left.
    """

    if len(self._intent_queue) > 0:
      intent = self._intent_queue.pop(0)
      self._cur_slots = intent['slots']
      self.go(intent['intent'])

  def _is_valid_request(self, data):
    """Checks if a request is valid in the current context.

    :param data: Request data
    :type data: dict

    """

    conversation_id = data.get(CID_KEY)

    if conversation_id == None:
      return False

    return conversation_id == self._cur_conversation_id

  def ask(self, data, raw_msg):
    """Ask required by the skill intent.

    :param data: Data sent by the intent, it should at least contains the key "slot"
    :type data: dict
    :param raw_msg: Raw payload received
    :type raw_msg: str

    """

    if self._is_valid_request(data):
      self.go(to_ask_state(data['slot']), payload=raw_msg)

  def show(self, data, raw_msg):
    """Show simply pass message to the client channel.

    :param data: Data sent by the intent
    :type data: dict
    :param raw_msg: Raw payload received
    :type raw_msg: str

    """

    if self._is_valid_request(data):
      self._client.show(raw_msg)

  def _terminate_from_skill(self, data, raw_msg):
    """Called by a skill when it has ended its work.

    :param data: Data sent by the intent
    :type data: dict
    :param raw_msg: Raw payload received
    :type raw_msg: str

    """

    if self._is_valid_request(data):
      self.terminate()

  def terminate(self):
    """Terminates a dialog and returns to the asleep state.
    """

    self.go(STATE_ASLEEP)

  def start(self, config):
    """Starts this agent.

    :param config: Broker configuration
    :type config: BrokerConfig

    """

    self._client.start(config)

  def cleanup(self):
    """Cleanup the agent.
    """

    # TODO Find a way to remove it cleanly
    # self._machine.remove_model(self)

    self._client.destroyed()
    self._client.stop()

  def __str__(self):
    return 'Agent %s - %s' % (self.id, self.interpreter.lang())



from setuptools import setup

setup(
  name='atlas-core',
  version='1.0.0',
  description='An open-source assistant built for people',
  author='Julien LEICHER',
  license='GPL-3.0',
  packages=['atlas', 'atlas.client', 'atlas.interpreters', 'atlas.web'],
  include_package_data=True,
  install_requires=[
    'atlas-sdk==1.0.0',
    'transitions==0.6.4',
    'Flask==0.12.2',
    'Flask-RESTful==0.3.6',
    'Flask-SocketIO==2.9.6',
    'PyYAML==3.12',
  ],
  entry_points={
    'console_scripts': [
      'atlas = atlas.cli:main',
      'atlas-check = atlas.cli:check',
      'atlas-web = atlas.cli:web',
      'atlas-client = atlas.cli:client',
    ]
  },
)


from .client import DiscoveryClient
from atlas_sdk import BrokerConfig
from .utils import find
from datetime import datetime, timedelta
import logging, threading

class Skill():
  """Represents a skill discovered by atlas.
  """

  def __init__(self, name, version, author=None, description=None, intents={}, env={}):
    """Constructs a new skill.
    
    :param name: Name of the skill
    :type name: str
    :param version: Version of the skill
    :type version: str
    :param author: Author of the skill
    :type author: str
    :param description: Optional description of the skill
    :type description: str
    :param intents: Intents supported by this skill with associated slots
    :type intents: dict
    :param env: Configuration variables with their types
    :type env: dict

    """

    self.name = name
    self.author = author
    self.version = version
    self.description = description
    self.intents = intents
    self.env = env
    
    self.heard_of()

  def heard_of(self):
    """Inform that this skill has just responded to the discovery challenge.
    """

    self.last_heard_of = datetime.utcnow()

  def has_timedout(self, date, timeout):
    return self.last_heard_of < date + timedelta(seconds = -timeout)

  def __str__(self):
    return 'Skill "%s"' % self.name

class DiscoveryConfig():
  """Represents the discovery configuration;
  """

  def __init__(self, interval=10, timeout=30):
    """Constructs a new configuration for the discovery service.

    :param interval: Polling interval in seconds
    :type interval: int
    :param timeout: Timeout in seconds, when a skill should be removed
    :type timeout: int

    """

    self.interval = interval
    self.timeout = timeout

class Discovery():
  """Represents the discovery service used to keep track of registered skills.
  """

  def __init__(self, config):
    """Constructs a new Discovery handler.

    :param config: Service configuration
    :type config: DiscoveryConfig
    """

    self._config = config
    self._thread = None
    self._skills = {}
    self._log = logging.getLogger('atlas.discovery')
    self._client = DiscoveryClient(on_discovery=self.process)

  def _ping(self):
    self._log.debug('Sending discovery request!')

    try:
      self._client.ping()
    except Exception as e:
      # May be because the client is not connected yet
      self._log.warn('Could not send a ping discovery: %s' % e)

    # Check for timeout skills

    now = datetime.utcnow()

    for k in list(self._skills.keys()):
      v = self._skills.get(k)

      if v.has_timedout(now, self._config.timeout):
        del self._skills[k]
        self._log.info('⏳ SKill %s has timed out, removed it' % k)

    self._thread = threading.Timer(self._config.interval, self._ping)
    self._thread.daemon = True
    self._thread.start()

  def process(self, skill_data, raw=None):
    """Process an incoming ping response for skill data.

    :param skill_data: Skill data received
    :type skill_data: dict
    :param raw: Raw message data
    :type raw: str

    """

    self._log.debug('Processing response from %s' % skill_data)

    name = skill_data.get('name')

    if name:
      skill = self._skills.get(name)

      if skill:
        self._log.debug('Skill %s already exists, updating' % name)
        skill.heard_of()
      else:
        self._skills[name] = Skill(**skill_data)
        self._log.info('🛠️ Added skill %s' % name)
    else:
      self._log.warn('No name defined, skipping the skill')

  def skill_env_for_intent(self, intent):
    """Try to retrieve a skill env associated with the intent name given.

    :param intent: Name of the intent
    :type intent: str
    :rtype: list

    """

    # TODO what if multiple skill can respond to the same intent with different env? :/

    skill = find(self._skills.values(), lambda s: s.intents.get(intent) != None)

    if skill:
      return skill.env.keys()

    return None

  def start(self, broker_config):
    """Starts the discovery service.

    :param broker_config: Broker configuration
    :type broker_config: BrokerConfig
    
    """

    self._client.start(broker_config)
    self._ping()

  def cleanup(self):
    """Cleanup discovery service stuff.
    """
    
    if self._thread:
      self._thread.cancel()

    self._client.stop()


from atlas_sdk.client import Client, \
  CHANNEL_CREATE_TOPIC, CHANNEL_DESTROY_TOPIC
import re, json

def split_client_id_and_operation(topic):
  """Split the topic to get the client_id and the operation name.

  :param topic: Message topic
  :type topic: str
  :rtype: str
  
  """

  r = re.search('atlas/(.*?)/channel/(.*)', topic)

  return (r.group(1), r.group(2))

class AtlasClient(Client):
  """MQTT client used to create and destroy Agent upon channel creation / destruction.

  This one is special.

  """

  def __init__(self, on_create=None, on_destroy=None):
    """Constructs a new AtlasClient.

    :param on_create: Handler when a channel has been created
    :type on_create: callable
    :param on_destroy: Handler when a channel has been destroyed
    :type on_destroy: callable

    """

    super(AtlasClient, self).__init__(name='atlas')

    self.on_create = on_create or self.handler_not_set
    self.on_destroy = on_destroy or self.handler_not_set

  def on_connect(self, client, userdata, flags, rc):
    super(AtlasClient, self).on_connect(client, userdata, flags, rc)

    client.subscribe(CHANNEL_CREATE_TOPIC % '+')
    client.subscribe(CHANNEL_DESTROY_TOPIC % '+')

  def on_message(self, client, userdata, msg):
    id, op = split_client_id_and_operation(msg.topic)

    if op == 'create':
      try:
        data = json.loads(msg.payload)
      except json.decoder.JSONDecodeError:
        data = {}
        self.log.warn('Could not decode payload %s' % msg.payload)

      data.update({ 'id': id })

      self.on_create(data)
    elif op == 'destroy':
      self.on_destroy(id)